#!/usr/bin/perl
use warnings;
use strict;
use English;
require Tk;
use Tk;
use Tk::FileSelect;
require Tk::FileSelect;
use Tk::ErrorDialog;
require Tk::ErrorDialog;

#############################################
# WORDSMITH.PL
# A Rhyming Dictionary
# (c) Amelie Zapf 2006
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place - Suite 330, Boston, MA 02111-1307, USA.
#############################################

our $version="0.1";
my $main=MainWindow->new();
$main->minsize(qw(500 300));
$main->title("Wordsmith Rhyming Dictionary Version $version");
$main->configure(-background=>'grey');
my $menu_bar=$main->Frame(-relief=>'groove',
			  -borderwidth=>3,
			  -background=>'grey'
			  )->pack(-side=>'top',
				  -fill=>'x');
my $file_mb=$menu_bar->Menubutton(-text=>'File',
				  -background=>'grey',
				  -activebackground=>'lightgrey',
				  -foreground=>'black'
				  )->pack(-side=>'left');
$file_mb->command(-label=>'Open Dictionary',
		  -activebackground=>'lightgrey',
		  -command=>\&load);
$file_mb->separator();
$file_mb->command(-label=>'Quit',
		  -activebackground=>'lightgrey',
		  -command=>sub{$main->destroy});
$file_mb->separator();
my $help_mb=$menu_bar->Menubutton(-text=>'Help',
				  -background=>'grey',
				  -activebackground=>'lightgrey',
				  -foreground=>'black'
				  )->pack(-side=>'right');
$help_mb->command(-label=>'Manual',
		  -activebackground=>'lightgrey',
		  -command=>\&manual_txt);
$help_mb->command(-label=>'About',
		  -activebackground=>'lightgrey',
		  -command=>\&about_txt);
$help_mb->separator(-background=>'grey');
my $top=$main->Frame(-background=>'grey'
		     )->pack(-side=>'top', -fill=>'x');
my $left1=$top->Frame(-background=>'grey'
		      )->pack(-side=>'left',
			      -pady=>8, -padx=>8);
my $enterlabel=$left1->Label(-text=>'Enter search word (regexp): ',
			     -background=>'grey'
			     )->pack(-anchor=>'w');
my $keyword;
my @indices;
my $enterfield=$left1->Entry(-background=>'white',
			     -textvariable=>\$keyword
			     )->pack(-anchor=>'w');
$enterfield->bind("<Return>", sub{&find($keyword)});
my $enterbutton=$left1->Button(-text=>'Submit',
			       -background=>'grey',
			       -foreground=>'black',
			       -activebackground=>'lightgrey',
			       -command=>sub{&find($keyword)}
			       )->pack(-anchor=>'e',
				       -pady=>10);
my $hitslabel=$left1->Label(-text=>'Hits: ',
			     -background=>'grey'
			     )->pack(-anchor=>'w');
my $hitsfield=$left1->Scrolled('Listbox', -height=>10,
			      -background=>'white',
			      -foreground=>'black',
			      -highlightcolor=>'blue',
			      -scrollbars=>'e'
			      )->pack(-anchor=>'w',
				      -fill=>'both');
$hitsfield->bind("<Button-1>", sub{&get_spelling()});
my @rhymetype;
our $rhymetype;
my $type;
my $radioright=$top->Frame(-background=>'grey'
		       )->pack(-side=>'left',
			       -pady=>8, -padx=>8,
			       -fill=>'y');
# what the hell is going wrong, ugly workaround
$rhymetype[0]=$radioright->Radiobutton(-value=>'end',
	  -variable=>\$rhymetype,
	  -background=>'grey',
	  -highlightthickness=>0,
	  -text=>'End Rhyme'
	  )->pack(side=>'top', -anchor=>'w');
$rhymetype[1]=$radioright->Radiobutton(-value=>'1',
          -variable=>\$rhymetype,
	  -background=>'grey',
	  -highlightthickness=>0,
	  -text=>'Force Masculine Rhyme'
          )->pack(side=>'top', -anchor=>'w');
$rhymetype[2]=$radioright->Radiobutton(-value=>'2',
	  -variable=>\$rhymetype,
	  -background=>'grey',
	  -highlightthickness=>0,
	  -text=>'Force Feminine Rhyme'
	  )->pack(side=>'top', -anchor=>'w');
$rhymetype[3]=$radioright->Radiobutton(-value=>'3',
	  -variable=>\$rhymetype,
	  -background=>'grey',
	  -highlightthickness=>0,
	  -text=>'Force 3-syllable Rhyme'
	  )->pack(side=>'top', -anchor=>'w');
$rhymetype[4]=$radioright->Radiobutton(-value=>'beg',
	  -variable=>\$rhymetype,
	  -background=>'grey',
	  -highlightthickness=>0,
	  -text=>'Beginning Rhyme'
	  )->pack(side=>'top', -anchor=>'w');
$rhymetype[5]=$radioright->Radiobutton(-value=>'all',
	  -variable=>\$rhymetype,
	  -background=>'grey',
	  -highlightthickness=>0,
	  -text=>'Alliteration'
	  )->pack(side=>'top', -anchor=>'w');
$rhymetype[6]=$radioright->Radiobutton(-value=>'cus',
	  -variable=>\$rhymetype,
	  -background=>'grey',
	  -highlightthickness=>0,
	  -text=>'Custom Regexp:'
	  )->pack(side=>'top', -anchor=>'w');
my $regexp=$radioright->Entry(-background=>'white'
			     )->pack(-anchor=>'e', -pady=>8);
my ($igcomp, $ighyph);
my $ig_comp=$radioright->Checkbutton(-text=>'Ignore word compounds',
				     -background=>'grey',
				     -foreground=>'black',
				     -highlightthickness=>0,
				     -offvalue=>0,
				     -onvalue=>1,
				     -variable=>\$igcomp
				     )->pack(-side=>'top', -anchor=>'w');
$ig_comp->select;
my $ig_hyph=$radioright->Checkbutton(-text=>'Ignore hyphenated words',
				     -background=>'grey',
				     -foreground=>'black',
				     -highlightthickness=>0,
				     -offvalue=>0,
				     -onvalue=>1,
				     -variable=>\$ighyph
				     )->pack(-side=>'top', -anchor=>'w');
$ig_hyph->deselect;
my $evalbutton=$radioright->Button(-text=>'Evaluate',
				     -background=>'grey',
				     -foreground=>'black',
				     -activebackground=>'lightgrey',
				     -command=>\&select
				     )->pack(-padx=>8, -pady=>8);
my $helpbutton=$radioright->Button(-text=>'Help on Regexps',
			       -background=>'grey',
			       -foreground=>'black',
			       -activebackground=>'lightgrey',
			       -command=>\&reghelp_txt
			      )->pack(-padx=>2, -pady=>2);
my $right1=$top->Frame(-background=>'grey'
		       )->pack(-side=>'right',
			       -pady=>8, -padx=>8);
my $resultslabel=$right1->Label(-text=>'Results: ',
			     -background=>'grey'
			     )->pack(-anchor=>'w',
				     -side=>'top');
my $resultsfield=$right1->Scrolled('Listbox', -height=>15,
			      -background=>'white',
			      -foreground=>'black',
			      -highlightcolor=>'blue',
			      -scrollbars=>'e'
			      )->pack(-anchor=>'w',
				      -fill=>'both');

   
our @words=();
our @pron=();
MainLoop();

sub manual_txt {
    my $text=<<'EOF';
Wordsmith was created to be used with the Moby Pronunciator, as obtainable 
from http://www.dcs.shef.ac.uk/research/ilash/Moby/. Any other dictionary
in the same format should also work fine.

If there is demand to also create phonetic word lists for other languages,
please let me know at amy@ameliezapf.com.

The usage of Wordsmith is straightforward:

Load a dictionary using "File->Open Dictionary".

Type the rhyme word into the box labeled "Enter Search Word" (this box also
can take Perl regular expressions to simplify your search). Hit "Submit"
to scan for words containing your entry in the dictionary. Click on one of
these, select the type of rhyme you want from the radiobutton list, and
choose if you'd like to retrieve single words only, or also compounds,
hyphenated and separate. Hit "Evaluate", and the results will be displayed
in the "Results" column. You're done.

The supported rhyme types are:

End Rhyme: rhyming from the last stressed syllable to the end of the word.
Masculine Rhyme:  Only the last syllable must rhyme.
Feminine Rhyme:   The two last syllables must rhyme.
3-syllable Rhyme: The three last syllables must rhyme.
Beginning Rhyme:  Words equal in pronunciation up until, and including,
                  the first vowel.
Alliteration:     Same consonant preceding the first stressed syllable.
Custom Regexp:    Free phonetic pattern search in the corpus.
EOF
    my $dialogBox=$main->DialogBox(-title=>'The Wordsmith Manual',
				   -default_button=>'OK',
				   -background=>'grey');
    my $dialogText=$dialogBox->Message(-text=>$text,
				       -background=>'grey')->pack();
    $dialogBox->Show();
}

sub reghelp_txt {
    my $text=<<'EOF';
The regular expressions used here are standard Perl, see 
www.perl.com/doc/manual/html/pod/perlre.html for details.

Additionally, C substitutes for any consonant, V for any vowel.
Consonants and vowels are separated from one another by slashes.

The phonetic nomenclature is for vowels:

&: a as in "dab", (@): ai as in "air", A: a as in "far", eI: ay as in "day",
@: schwa as a in "ado", -: schwa as dl in "handle",  E: e as in "red", 
i: ee as in "see", I: i as in "hid", aI: i as in "ice", Oi: oi as in "oil", 
AU: ow as in "how", O: o as in "dog", oU: oa as in "boat", u: oo as in "too", 
U: oo as in "book", @r: ur as in "burn".

For consonants:

b: b as in "nab", tS: ch as in "ouch", d: d as in "pod", f: f as in "elf", 
g: g as in "fig", h: h as in "had", hw: wh as in "white", 
dz: g as in "vegetable", k: c as in "act", l: l as in "ail", m: m as in "aim", 
N: ng as in "bang", n: n as in "and", p: p as in "imp", r: r as in "ire",
S: sh as in "she", s: s as in "sip", T: th as in "bath", D: th as in "the",
t: t as in "tap", v: v as in "average", w: w as in "win", j: y as in "you", 
Z: s as in "vision", z: z as in "zoo".

Primary stress is denoted ', secondary stress ,.
EOF
    my $dialogBox=$main->DialogBox(-title=>'Help on Regular Expressions',
				   -default_button=>'OK',
				   -background=>'grey');
    my $dialogText=$dialogBox->Message(-text=>$text,
				       -background=>'grey')->pack();
    $dialogBox->Show();
}

sub about_txt {
    my $text=<<EOF;
Wordsmith Rhyming Dictionary Version $version

This is ALPHA code! Please report bugs to amy\@ameliezapf.com.

Written by Amelie Zapf.
http://www.ameliezapf.com

(c) Amelie Zapf 2006.
EOF
    my $dialogBox=$main->DialogBox(-title=>'About Wordsmith Rhyming Dictionary',
				   -default_button=>'OK',
				   -background=>'grey');
    my $dialogText=$dialogBox->Message(-text=>$text,
				       -background=>'grey')->pack();
    $dialogBox->Show();
}


sub select {
    my $rhymesound=$regexp->get;
    $rhymesound=&trim($rhymesound, $rhymetype);
    my @rhymes;
    my @newrhymes=();
    for (my $i=0; $i<scalar @pron; $i++) {
	my $tmp=$pron[$i];
	if ($tmp!~/[,']/) { $tmp='\''.$tmp; }
	if ($tmp=~m|$rhymesound|) {
	    push @rhymes, ($i);
	}
    }
    foreach (@rhymes) { 
        $_=$words[$_];
        unless (($ighyph && /-/) || ($igcomp && / /)) {
            push @newrhymes, ($_);
        }
    }
    unshift @newrhymes, ('end');
    $resultsfield->delete(0, 'end');
    $resultsfield->insert(@newrhymes);
}

sub trim {
    my ($sound, $type)=@_;
    # highly crufty & bodacious
    my @vowels=("\\(\@\\)", "\"Y\"", "\"A\"", "eI", "aI", "Oi", "AU", "oU", 
		# '\@r',
		"&",  "A", "\\@", "-", "E", "i", "I", "A", "O", "u", "U", "y");
    my @consonants=("\"N\"", "\"R\"", "tS", "hw", "dZ", "b", "d", "f", "g", 
		    "h", "k", "l", "m", "N", "n", "p", "r", "S", "s", "T", 
		    "D", "t", "v", "w", "j", "Z", "z", "x", "y");
    my $anyvow=join ")(", @vowels;
    $anyvow="(".$anyvow.")";
    my $anycon=join ")(", @consonants;
    $anycon="(".$anycon.")";
   
    if ($type eq "end") {
	# for end rhyme: split into stresses
	my @tmp=split /[ ',]/, $sound;
        $sound=pop @tmp;
        # eliminate leading consonants
        $sound=~s/^\^//;
        $sound=~s/^[\/$anycon]*//;
        $sound="[ ',][\/$anycon]*".$sound;
    } elsif ($type eq "1" || $type eq "2" || $type eq "3") {
        # forced-length rhymes: split into phonemes
        $sound=~s/^\^//;
        my @tmp=split /\//, $sound;
        foreach (@tmp) { if (/^[$anyvow]/) { $_="__" . $_; } }
        $sound=join "/", @tmp;
        @tmp=split /__/, $sound;
        shift @tmp;
        my @tmp2=();
        $sound="";
        if ($type>scalar @tmp) { 
            my $dialogBox=$main->DialogBox(-title=>'Warning',
				   -default_button=>'OK',
				   -background=>'grey');
            my $dialogText=$dialogBox->Message(
                 -text=>'Not enough syllables for this rhyme',
		 -background=>'grey')->pack();
            $dialogBox->Show();
        } else {
            for (my $i=0; $i<$type; $i++) {
                my $tmp=pop @tmp;
                unshift @tmp2, ($tmp);
            }
            $sound=join "", @tmp2;
        }
    } elsif ($type eq "beg") {
        # Beginning rhyme is identicity of the first syllable, up to the
        # vowel.
        $sound=~s/^\^//;
        my @tmp=split /\//, $sound;
        my @tmp2=();
        for (my $i=0; $i<scalar @tmp; $i++) {
            $tmp2[$i]=$tmp[$i];
            if ($tmp[$i]=~/^[,']?[$anyvow]/) { last; }
        }
        $sound=join "/", @tmp2;
        $sound=~s/^[,']//;
        $sound="^[^,']*".$sound;
    } elsif ($type eq "all") {
        # Alliteration is identicity of the first stressed phoneme
        $sound=~s/^\^//;
        my @tmp=split /\//, $sound;
        if ($sound=~/[',]/) { while ($tmp[0]!~/[',]/) { shift @tmp; } }
        $tmp[0]=~s/[',]//;
        $sound="^[^,']*[,']".$tmp[0];
    } elsif ($type eq "cus") {
        $sound=~s/V/[$anyvow]/g;
	$sound=~s/C/[$anycon]/g;
    }
    return $sound;
}

sub get_spelling {
    my $word=$hitsfield->curselection;
    $word=$indices[++$word];
    for (my $i=0; $i<scalar @words; $i++) {
	if ($word eq $words[$i]) {
	    # Perl-regexp it
	    my $tmp=$pron[$i];
	    $tmp=~s/\\/\\\\/g;
	    $tmp=~s/\|/\\|/g;
	    $tmp=~s/\(/\\(/g;
	    $tmp=~s/\)/\\)/g;
	    $tmp=~s/\[/\\[/g;
	    $tmp=~s/\{/\\{/g;
	    $tmp=~s/\^/\\^/g;
	    $tmp=~s/\$/\\\$/g;
	    $tmp=~s/\*/\\*/g;
	    $tmp=~s/\+/\\+/g;
	    $tmp=~s/\?/\\?/g;
	    $tmp=~s/\./\\./g;
	    $tmp="^".$tmp."\$";
	    $regexp->delete(0, 'end');
	    $regexp->insert(0, $tmp);
	    last;
	}
    }
    return 1;
}

sub load {
    my $file="";
    while (!$file) {
	my $dialog=$main->FileSelect(-filter=>'*.unc',
				     -directory=>'.');
	$file=$dialog->Show;
    }
    open FILE, $file or die "Couldn't open file $file.\n";
    my @content=<FILE>;
    close FILE;
    @words=();
    @pron=();
    my $content=join "", @content;
    @content=split /\r/, $content;
    for (my $i=0; $i<scalar @content; $i++) {
	my @pair=split / /, $content[$i];
	$pair[0]=~s/_/ /g;
	$pair[1]=~s/_/ /g;
	$pair[1]=~s/'/\/'/g;
	$pair[1]=~s/,/\/,/g;
	$pair[1]=~s/\/+/\//g;
	$pair[1]=~s/'\//'/g;
	$pair[1]=~s/,\//,/g;
	$pair[1]=~s/^\/+//;
	$pair[1]=~s/\/+$//;
	push @words, ($pair[0]);
	push @pron, ($pair[1]);
    }
    return 1;
}
	
sub find {
    my $word=$_[0];
    @indices=();
    for (my $i=0; $i<scalar @words; $i++) {
	if ($words[$i]=~/$word/i) { push @indices, ($words[$i]); }
    }
    unshift @indices, ('end');
    $hitsfield->delete(0, 'end');
    $hitsfield->insert(@indices);
}
